@@ -10,7 +10,6 @@
 package cache
 
 import (
-	"bytes"
 	"context"
 	"encoding/json"
 	"errors"
@@ -51,7 +50,7 @@ type Folder struct {
 	Dir     protocol.DocumentURI
 	Name    string // decorative name for UI; not necessarily unique
 	Options *settings.Options
-	Env     *GoEnv
+	Env     GoEnv
 }
 
 // GoEnv holds the environment variables and data from the Go command that is
@@ -73,8 +72,19 @@ type GoEnv struct {
 	GoVersionOutput string // complete go version output
 
 	// OS environment variables (notably not go env).
-	GOWORK           string
-	GOPACKAGESDRIVER string
+
+	// ExplicitGOWORK is the GOWORK value set explicitly in the environment. This
+	// may differ from `go env GOWORK` when the GOWORK value is implicit from the
+	// working directory.
+	ExplicitGOWORK string
+
+	// EffectiveGOPACKAGESDRIVER is the effective go/packages driver binary that
+	// will be used. This may be set via GOPACKAGESDRIVER, or may be discovered
+	// via os.LookPath("gopackagesdriver"). The latter functionality is
+	// undocumented and may be removed in the future.
+	//
+	// If GOPACKAGESDRIVER is set to "off", EffectiveGOPACKAGESDRIVER is "".
+	EffectiveGOPACKAGESDRIVER string
 }
 
 // View represents a single build for a workspace.
@@ -203,7 +213,7 @@ func (d *viewDefinition) GOOS() string {
 	return d.folder.Env.GOOS
 }
 
-// GOOS returns the effective GOARCH value for this view definition, accounting
+// GOARCH returns the effective GOARCH value for this view definition, accounting
 // for its env overlay.
 func (d *viewDefinition) GOARCH() string {
 	if goarch, ok := d.envOverlay["GOARCH"]; ok {
@@ -302,23 +312,23 @@ const (
 func (t ViewType) String() string {
 	switch t {
 	case GoPackagesDriverView:
-		return "GoPackagesDriverView"
+		return "GoPackagesDriver"
 	case GOPATHView:
-		return "GOPATHView"
+		return "GOPATH"
 	case GoModView:
-		return "GoModView"
+		return "GoMod"
 	case GoWorkView:
-		return "GoWorkView"
+		return "GoWork"
 	case AdHocView:
-		return "AdHocView"
+		return "AdHoc"
 	default:
 		return "Unknown"
 	}
 }
 
-// moduleMode reports whether the view uses Go modules.
-func (w viewDefinition) moduleMode() bool {
-	switch w.typ {
+// usesModules reports whether the view uses Go modules.
+func (typ ViewType) usesModules() bool {
+	switch typ {
 	case GoModView, GoWorkView:
 		return true
 	default:
@@ -326,50 +336,12 @@ func (w viewDefinition) moduleMode() bool {
 	}
 }
 
+// ID returns the unique ID of this View.
 func (v *View) ID() string { return v.id }
 
-// tempModFile creates a temporary go.mod file based on the contents
-// of the given go.mod file. On success, it is the caller's
-// responsibility to call the cleanup function when the file is no
-// longer needed.
-func tempModFile(modURI protocol.DocumentURI, gomod, gosum []byte) (tmpURI protocol.DocumentURI, cleanup func(), err error) {
-	filenameHash := file.HashOf([]byte(modURI.Path()))
-	tmpMod, err := os.CreateTemp("", fmt.Sprintf("go.%s.*.mod", filenameHash))
-	if err != nil {
-		return "", nil, err
-	}
-	defer tmpMod.Close()
-
-	tmpURI = protocol.URIFromPath(tmpMod.Name())
-	tmpSumName := sumFilename(tmpURI)
-
-	if _, err := tmpMod.Write(gomod); err != nil {
-		return "", nil, err
-	}
-
-	// We use a distinct name here to avoid subtlety around the fact
-	// that both 'return' and 'defer' update the "cleanup" variable.
-	doCleanup := func() {
-		_ = os.Remove(tmpSumName)
-		_ = os.Remove(tmpURI.Path())
-	}
-
-	// Be careful to clean up if we return an error from this function.
-	defer func() {
-		if err != nil {
-			doCleanup()
-			cleanup = nil
-		}
-	}()
-
-	// Create an analogous go.sum, if one exists.
-	if gosum != nil {
-		if err := os.WriteFile(tmpSumName, gosum, 0655); err != nil {
-			return "", nil, err
-		}
-	}
-
-	return tmpURI, doCleanup, nil
+// GoCommandRunner returns the shared gocommand.Runner for this view.
+func (v *View) GoCommandRunner() *gocommand.Runner {
+	return v.gocmdRunner
 }
 
 // Folder returns the folder at the base of this view.
@@ -407,32 +379,6 @@ func (s *Session) UpdateFolders(ctx context.Context, newFolders []*Folder) error
 	return nil
 }
 
-// viewEnv returns a string describing the environment of a newly created view.
-//
-// It must not be called concurrently with any other view methods.
-// TODO(rfindley): rethink this function, or inline sole call.
-func viewEnv(v *View) string {
-	var buf bytes.Buffer
-	fmt.Fprintf(&buf, `go info for %v
-(view type %v)
-(root dir %s)
-(go version %s)
-(build flags: %v)
-(go env: %+v)
-(env overlay: %v)
-`,
-		v.folder.Dir.Path(),
-		v.typ,
-		v.root.Path(),
-		strings.TrimRight(v.folder.Env.GoVersionOutput, "\n"),
-		v.folder.Options.BuildFlags,
-		*v.folder.Env,
-		v.envOverlay,
-	)
-
-	return buf.String()
-}
-
 // RunProcessEnvFunc runs fn with the process env for this snapshot's view.
 // Note: the process env contains cached module and filesystem state.
 func (s *Snapshot) RunProcessEnvFunc(ctx context.Context, fn func(context.Context, *imports.Options) error) error {
@@ -894,9 +840,9 @@ func defineView(ctx context.Context, fs file.Source, folder *Folder, forFile fil
 	var err error
 	dirURI := protocol.URIFromPath(dir)
 	goworkFromEnv := false
-	if folder.Env.GOWORK != "off" && folder.Env.GOWORK != "" {
+	if folder.Env.ExplicitGOWORK != "off" && folder.Env.ExplicitGOWORK != "" {
 		goworkFromEnv = true
-		def.gowork = protocol.URIFromPath(folder.Env.GOWORK)
+		def.gowork = protocol.URIFromPath(folder.Env.ExplicitGOWORK)
 	} else {
 		def.gowork, err = findRootPattern(ctx, dirURI, "go.work", fs)
 		if err != nil {
@@ -920,20 +866,10 @@ func defineView(ctx context.Context, fs file.Source, folder *Folder, forFile fil
 	//  - def.envOverlay.
 
 	// If GOPACKAGESDRIVER is set it takes precedence.
-	{
-		// The value of GOPACKAGESDRIVER is not returned through the go command.
-		gopackagesdriver := os.Getenv("GOPACKAGESDRIVER")
-		// A user may also have a gopackagesdriver binary on their machine, which
-		// works the same way as setting GOPACKAGESDRIVER.
-		//
-		// TODO(rfindley): remove this call to LookPath. We should not support this
-		// undocumented method of setting GOPACKAGESDRIVER.
-		tool, err := exec.LookPath("gopackagesdriver")
-		if gopackagesdriver != "off" && (gopackagesdriver != "" || (err == nil && tool != "")) {
-			def.typ = GoPackagesDriverView
-			def.root = dirURI
-			return def, nil
-		}
+	if def.folder.Env.EffectiveGOPACKAGESDRIVER != "" {
+		def.typ = GoPackagesDriverView
+		def.root = dirURI
+		return def, nil
 	}
 
 	// From go.dev/ref/mod, module mode is active if GO111MODULE=on, or
@@ -959,7 +895,7 @@ func defineView(ctx context.Context, fs file.Source, folder *Folder, forFile fil
 
 	// Prefer a go.work file if it is available and contains the module relevant
 	// to forURI.
-	if def.adjustedGO111MODULE() != "off" && folder.Env.GOWORK != "off" && def.gowork != "" {
+	if def.adjustedGO111MODULE() != "off" && folder.Env.ExplicitGOWORK != "off" && def.gowork != "" {
 		def.typ = GoWorkView
 		if goworkFromEnv {
 			// The go.work file could be anywhere, which can lead to confusing error
@@ -1062,18 +998,20 @@ func FetchGoEnv(ctx context.Context, folder protocol.DocumentURI, opts *settings
 
 	// The value of GOPACKAGESDRIVER is not returned through the go command.
 	if driver, ok := opts.Env["GOPACKAGESDRIVER"]; ok {
-		env.GOPACKAGESDRIVER = driver
-	} else {
-		env.GOPACKAGESDRIVER = os.Getenv("GOPACKAGESDRIVER")
+		if driver != "off" {
+			env.EffectiveGOPACKAGESDRIVER = driver
+		}
+	} else if driver := os.Getenv("GOPACKAGESDRIVER"); driver != "off" {
+		env.EffectiveGOPACKAGESDRIVER = driver
 		// A user may also have a gopackagesdriver binary on their machine, which
 		// works the same way as setting GOPACKAGESDRIVER.
 		//
 		// TODO(rfindley): remove this call to LookPath. We should not support this
 		// undocumented method of setting GOPACKAGESDRIVER.
-		if env.GOPACKAGESDRIVER == "" {
+		if env.EffectiveGOPACKAGESDRIVER == "" {
 			tool, err := exec.LookPath("gopackagesdriver")
 			if err == nil && tool != "" {
-				env.GOPACKAGESDRIVER = tool
+				env.EffectiveGOPACKAGESDRIVER = tool
 			}
 		}
 	}
@@ -1082,9 +1020,9 @@ func FetchGoEnv(ctx context.Context, folder protocol.DocumentURI, opts *settings
 	// between an explicit GOWORK value and one which is implicit from the file
 	// system. The former doesn't change unless the environment changes.
 	if gowork, ok := opts.Env["GOWORK"]; ok {
-		env.GOWORK = gowork
+		env.ExplicitGOWORK = gowork
 	} else {
-		env.GOWORK = os.Getenv("GOWORK")
+		env.ExplicitGOWORK = os.Getenv("GOWORK")
 	}
 	return env, nil
 }
