@@ -11,19 +11,18 @@ import (
 	"go/ast"
 	"strings"
 
-	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/gopls/internal/analysis/fillstruct"
 	"golang.org/x/tools/gopls/internal/analysis/fillswitch"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/file"
+	"golang.org/x/tools/gopls/internal/label"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/protocol/command"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/slices"
 	"golang.org/x/tools/internal/event"
-	"golang.org/x/tools/internal/event/tag"
 	"golang.org/x/tools/internal/imports"
 )
 
@@ -33,8 +32,17 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 	// Only compute quick fixes if there are any diagnostics to fix.
 	wantQuickFixes := want[protocol.QuickFix] && len(diagnostics) > 0
 
-	// Code actions requiring syntax information alone.
-	if wantQuickFixes || want[protocol.SourceOrganizeImports] || want[protocol.RefactorExtract] {
+	// Note: don't forget to update the allow-list in Server.CodeAction
+	// when adding new query operations like GoTest and GoDoc that
+	// are permitted even in generated source files
+
+	// Code actions that can be offered based on syntax information alone.
+	if wantQuickFixes ||
+		want[protocol.SourceOrganizeImports] ||
+		want[protocol.RefactorExtract] ||
+		want[protocol.GoDoc] ||
+		want[protocol.GoFreeSymbols] {
+
 		pgf, err := snapshot.ParseGo(ctx, fh, parsego.Full)
 		if err != nil {
 			return nil, err
@@ -44,7 +52,7 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 		if wantQuickFixes || want[protocol.SourceOrganizeImports] {
 			importEdits, importEditsPerFix, err := allImportsFixes(ctx, snapshot, pgf)
 			if err != nil {
-				event.Error(ctx, "imports fixes", err, tag.File.Of(fh.URI().Path()))
+				event.Error(ctx, "imports fixes", err, label.File.Of(fh.URI().Path()))
 				importEdits = nil
 				importEditsPerFix = nil
 			}
@@ -60,9 +68,8 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 					actions = append(actions, protocol.CodeAction{
 						Title: importFixTitle(importFix.fix),
 						Kind:  protocol.QuickFix,
-						Edit: &protocol.WorkspaceEdit{
-							DocumentChanges: documentChanges(fh, importFix.edits),
-						},
+						Edit: protocol.NewWorkspaceEdit(
+							protocol.DocumentChangeEdit(fh, importFix.edits)),
 						Diagnostics: fixed,
 					})
 				}
@@ -74,9 +81,8 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 				actions = append(actions, protocol.CodeAction{
 					Title: "Organize Imports",
 					Kind:  protocol.SourceOrganizeImports,
-					Edit: &protocol.WorkspaceEdit{
-						DocumentChanges: documentChanges(fh, importEdits),
-					},
+					Edit: protocol.NewWorkspaceEdit(
+						protocol.DocumentChangeEdit(fh, importEdits)),
 				})
 			}
 		}
@@ -88,13 +94,38 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 			}
 			actions = append(actions, extractions...)
 		}
+
+		if want[protocol.GoDoc] {
+			loc := protocol.Location{URI: pgf.URI, Range: rng}
+			cmd, err := command.NewDocCommand("View package documentation", loc)
+			if err != nil {
+				return nil, err
+			}
+			actions = append(actions, protocol.CodeAction{
+				Title:   cmd.Title,
+				Kind:    protocol.GoDoc,
+				Command: &cmd,
+			})
+		}
+
+		if want[protocol.GoFreeSymbols] && rng.End != rng.Start {
+			cmd, err := command.NewFreeSymbolsCommand("Show free symbols", pgf.URI, rng)
+			if err != nil {
+				return nil, err
+			}
+			// For implementation, see commandHandler.showFreeSymbols.
+			actions = append(actions, protocol.CodeAction{
+				Title:   cmd.Title,
+				Kind:    protocol.GoFreeSymbols,
+				Command: &cmd,
+			})
+		}
 	}
 
 	// Code actions requiring type information.
 	if want[protocol.RefactorRewrite] ||
 		want[protocol.RefactorInline] ||
-		want[protocol.GoTest] ||
-		want[protocol.GoDoc] {
+		want[protocol.GoTest] {
 		pkg, pgf, err := NarrowestPackageForFile(ctx, snapshot, fh.URI())
 		if err != nil {
 			return nil, err
@@ -122,19 +153,6 @@ func CodeActions(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle,
 			}
 			actions = append(actions, fixes...)
 		}
-
-		if want[protocol.GoDoc] {
-			loc := protocol.Location{URI: pgf.URI, Range: rng}
-			cmd, err := command.NewDocCommand("View package documentation", loc)
-			if err != nil {
-				return nil, err
-			}
-			actions = append(actions, protocol.CodeAction{
-				Title:   cmd.Title,
-				Kind:    protocol.GoDoc,
-				Command: &cmd,
-			})
-		}
 	}
 	return actions, nil
 }
@@ -205,7 +223,7 @@ func getExtractCodeActions(pgf *parsego.File, rng protocol.Range, options *setti
 	}
 	puri := pgf.URI
 	var commands []protocol.Command
-	if _, ok, methodOk, _ := CanExtractFunction(pgf.Tok, start, end, pgf.Src, pgf.File); ok {
+	if _, ok, methodOk, _ := canExtractFunction(pgf.Tok, start, end, pgf.Src, pgf.File); ok {
 		cmd, err := command.NewApplyFixCommand("Extract function", command.ApplyFixArgs{
 			Fix:          fixExtractFunction,
 			URI:          puri,
@@ -229,7 +247,7 @@ func getExtractCodeActions(pgf *parsego.File, rng protocol.Range, options *setti
 			commands = append(commands, cmd)
 		}
 	}
-	if _, _, ok, _ := CanExtractVariable(start, end, pgf.File); ok {
+	if _, _, ok, _ := canExtractVariable(start, end, pgf.File); ok {
 		cmd, err := command.NewApplyFixCommand("Extract variable", command.ApplyFixArgs{
 			Fix:          fixExtractVariable,
 			URI:          puri,
@@ -295,17 +313,17 @@ func getRewriteCodeActions(ctx context.Context, pkg *cache.Package, snapshot *ca
 		actions = append(actions, newCodeAction("Refactor: remove unused parameter", protocol.RefactorRewrite, &cmd, nil, options))
 	}
 
-	if action, ok := ConvertStringLiteral(pgf, fh, rng); ok {
-		actions = append(actions, action)
-	}
-
 	start, end, err := pgf.RangePos(rng)
 	if err != nil {
 		return nil, err
 	}
 
+	if action, ok := convertStringLiteral(pgf, fh, start, end); ok {
+		actions = append(actions, action)
+	}
+
 	var commands []protocol.Command
-	if _, ok, _ := CanInvertIfCondition(pgf.File, start, end); ok {
+	if _, ok, _ := canInvertIfCondition(pgf.File, start, end); ok {
 		cmd, err := command.NewApplyFixCommand("Invert 'if' condition", command.ApplyFixArgs{
 			Fix:          fixInvertIfCondition,
 			URI:          pgf.URI,
@@ -318,7 +336,7 @@ func getRewriteCodeActions(ctx context.Context, pkg *cache.Package, snapshot *ca
 		commands = append(commands, cmd)
 	}
 
-	if msg, ok, _ := CanSplitLines(pgf.File, pkg.FileSet(), start, end); ok {
+	if msg, ok, _ := canSplitLines(pgf.File, pkg.FileSet(), start, end); ok {
 		cmd, err := command.NewApplyFixCommand(msg, command.ApplyFixArgs{
 			Fix:          fixSplitLines,
 			URI:          pgf.URI,
@@ -331,7 +349,7 @@ func getRewriteCodeActions(ctx context.Context, pkg *cache.Package, snapshot *ca
 		commands = append(commands, cmd)
 	}
 
-	if msg, ok, _ := CanJoinLines(pgf.File, pkg.FileSet(), start, end); ok {
+	if msg, ok, _ := canJoinLines(pgf.File, pkg.FileSet(), start, end); ok {
 		cmd, err := command.NewApplyFixCommand(msg, command.ApplyFixArgs{
 			Fix:          fixJoinLines,
 			URI:          pgf.URI,
@@ -344,12 +362,10 @@ func getRewriteCodeActions(ctx context.Context, pkg *cache.Package, snapshot *ca
 		commands = append(commands, cmd)
 	}
 
-	// N.B.: an inspector only pays for itself after ~5 passes, which means we're
-	// currently not getting a good deal on this inspection.
-	//
-	// TODO: Consider removing the inspection after convenienceAnalyzers are removed.
-	inspect := inspector.New([]*ast.File{pgf.File})
-	for _, diag := range fillstruct.Diagnose(inspect, start, end, pkg.Types(), pkg.TypesInfo()) {
+	// fillstruct.Diagnose is a lazy analyzer: all it gives us is
+	// the (start, end, message) of each SuggestedFix; the actual
+	// edit is computed only later by ApplyFix, which calls fillstruct.SuggestedFix.
+	for _, diag := range fillstruct.Diagnose(pgf.File, start, end, pkg.Types(), pkg.TypesInfo()) {
 		rng, err := pgf.Mapper.PosRange(pgf.Tok, diag.Pos, diag.End)
 		if err != nil {
 			return nil, err
@@ -368,26 +384,15 @@ func getRewriteCodeActions(ctx context.Context, pkg *cache.Package, snapshot *ca
 		}
 	}
 
-	for _, diag := range fillswitch.Diagnose(inspect, start, end, pkg.Types(), pkg.TypesInfo()) {
-		edits, err := suggestedFixToEdits(ctx, snapshot, pkg.FileSet(), &diag.SuggestedFixes[0])
+	for _, diag := range fillswitch.Diagnose(pgf.File, start, end, pkg.Types(), pkg.TypesInfo()) {
+		changes, err := suggestedFixToDocumentChange(ctx, snapshot, pkg.FileSet(), &diag.SuggestedFixes[0])
 		if err != nil {
 			return nil, err
 		}
-
-		changes := []protocol.DocumentChanges{} // must be a slice
-		for _, edit := range edits {
-			edit := edit
-			changes = append(changes, protocol.DocumentChanges{
-				TextDocumentEdit: &edit,
-			})
-		}
-
 		actions = append(actions, protocol.CodeAction{
 			Title: diag.Message,
 			Kind:  protocol.RefactorRewrite,
-			Edit: &protocol.WorkspaceEdit{
-				DocumentChanges: changes,
-			},
+			Edit:  protocol.NewWorkspaceEdit(changes...),
 		})
 	}
 	for i := range commands {
@@ -411,33 +416,33 @@ func canRemoveParameter(pkg *cache.Package, pgf *parsego.File, rng protocol.Rang
 	if perrors, terrors := pkg.ParseErrors(), pkg.TypeErrors(); len(perrors) > 0 || len(terrors) > 0 {
 		return false // can't remove parameters from packages with errors
 	}
-	info, err := FindParam(pgf, rng)
+	info, err := findParam(pgf, rng)
 	if err != nil {
 		return false // e.g. invalid range
 	}
-	if info.Field == nil {
+	if info.field == nil {
 		return false // range does not span a parameter
 	}
-	if info.Decl.Body == nil {
+	if info.decl.Body == nil {
 		return false // external function
 	}
-	if len(info.Field.Names) == 0 {
+	if len(info.field.Names) == 0 {
 		return true // no names => field is unused
 	}
-	if info.Name == nil {
+	if info.name == nil {
 		return false // no name is indicated
 	}
-	if info.Name.Name == "_" {
+	if info.name.Name == "_" {
 		return true // trivially unused
 	}
 
-	obj := pkg.TypesInfo().Defs[info.Name]
+	obj := pkg.TypesInfo().Defs[info.name]
 	if obj == nil {
 		return false // something went wrong
 	}
 
 	used := false
-	ast.Inspect(info.Decl.Body, func(node ast.Node) bool {
+	ast.Inspect(info.decl.Body, func(node ast.Node) bool {
 		if n, ok := node.(*ast.Ident); ok && pkg.TypesInfo().Uses[n] == obj {
 			used = true
 		}
@@ -455,7 +460,7 @@ func getInlineCodeActions(pkg *cache.Package, pgf *parsego.File, rng protocol.Ra
 
 	// If range is within call expression, offer to inline the call.
 	var commands []protocol.Command
-	if _, fn, err := EnclosingStaticCall(pkg, pgf, start, end); err == nil {
+	if _, fn, err := enclosingStaticCall(pkg, pgf, start, end); err == nil {
 		cmd, err := command.NewApplyFixCommand(fmt.Sprintf("Inline call to %s", fn.Name()), command.ApplyFixArgs{
 			Fix:          fixInlineCall,
 			URI:          pgf.URI,
@@ -478,23 +483,21 @@ func getInlineCodeActions(pkg *cache.Package, pgf *parsego.File, rng protocol.Ra
 
 // getGoTestCodeActions returns any "run this test/benchmark" code actions for the selection.
 func getGoTestCodeActions(pkg *cache.Package, pgf *parsego.File, rng protocol.Range) ([]protocol.CodeAction, error) {
-	fns, err := TestsAndBenchmarks(pkg, pgf)
+	testFuncs, benchFuncs, err := testsAndBenchmarks(pkg.TypesInfo(), pgf)
 	if err != nil {
 		return nil, err
 	}
 
 	var tests, benchmarks []string
-	for _, fn := range fns.Tests {
-		if !protocol.Intersect(fn.Rng, rng) {
-			continue
+	for _, fn := range testFuncs {
+		if protocol.Intersect(fn.rng, rng) {
+			tests = append(tests, fn.name)
 		}
-		tests = append(tests, fn.Name)
 	}
-	for _, fn := range fns.Benchmarks {
-		if !protocol.Intersect(fn.Rng, rng) {
-			continue
+	for _, fn := range benchFuncs {
+		if protocol.Intersect(fn.rng, rng) {
+			benchmarks = append(benchmarks, fn.name)
 		}
-		benchmarks = append(benchmarks, fn.Name)
 	}
 
 	if len(tests) == 0 && len(benchmarks) == 0 {
@@ -511,7 +514,3 @@ func getGoTestCodeActions(pkg *cache.Package, pgf *parsego.File, rng protocol.Ra
 		Command: &cmd,
 	}}, nil
 }
-
-func documentChanges(fh file.Handle, edits []protocol.TextEdit) []protocol.DocumentChanges {
-	return protocol.TextEditsToDocumentChanges(fh.URI(), fh.Version(), edits)
-}
