@@ -25,6 +25,10 @@ package golang
 // - modify JS httpGET function to give a transient visual indication
 //   when clicking a source link that the editor is being navigated
 //   (in case it doesn't raise itself, like VS Code).
+// - move this into a new package, golang/pkgdoc, and then
+//   split out the various helpers without fear of polluting
+//   the golang package namespace.
+// - show "Deprecated" chip when appropriate.
 
 import (
 	"bytes"
@@ -36,8 +40,8 @@ import (
 	"go/token"
 	"go/types"
 	"html"
-	"log"
 	"path/filepath"
+	"strings"
 
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/protocol"
@@ -49,6 +53,15 @@ import (
 	"golang.org/x/tools/internal/typesinternal"
 )
 
+// TODO(adonovan): factor these two functions into an interface.
+type (
+	// A PkgURLFunc forms URLs of package or symbol documentation.
+	PkgURLFunc = func(path PackagePath, fragment string) protocol.URI
+
+	// A PosURLFunc forms URLs that cause the editor to navigate to a position.
+	PosURLFunc = func(filename string, line, col8 int) protocol.URI
+)
+
 // RenderPackageDoc formats the package documentation page.
 //
 // The posURL function returns a URL that when visited, has the side
@@ -57,7 +70,9 @@ import (
 //
 // The pkgURL function returns a URL for the documentation of the
 // specified package and symbol.
-func RenderPackageDoc(pkg *cache.Package, posURL func(filename string, line, col8 int) protocol.URI, pkgURL func(path PackagePath, fragment string) protocol.URI) ([]byte, error) {
+//
+// TODO(adonovan): "Render" is a client-side verb; rename to PackageDocHTML.
+func RenderPackageDoc(pkg *cache.Package, posURL PosURLFunc, pkgURL PkgURLFunc) ([]byte, error) {
 	// We can't use doc.NewFromFiles (even with doc.PreserveAST
 	// mode) as it calls ast.NewPackage which assumes that each
 	// ast.File has an ast.Scope and resolves identifiers to
@@ -160,9 +175,6 @@ func RenderPackageDoc(pkg *cache.Package, posURL func(filename string, line, col
 			return "", false
 		}
 		parser.LookupSym = func(recv, name string) (ok bool) {
-			defer func() {
-				log.Printf("LookupSym %q %q = %t ", recv, name, ok)
-			}()
 			// package-level decl?
 			if recv == "" {
 				return pkg.Types().Scope().Lookup(name) != nil
@@ -181,12 +193,19 @@ func RenderPackageDoc(pkg *cache.Package, posURL func(filename string, line, col
 		}
 	}
 
+	scope := pkg.Types().Scope()
+	escape := html.EscapeString
+
+	title := fmt.Sprintf("%s package - %s - Gopls packages",
+		pkg.Types().Name(), escape(pkg.Types().Path()))
+
 	var buf bytes.Buffer
 	buf.WriteString(`<!DOCTYPE html>
 <html>
 <head>
   <meta charset="UTF-8">
   <style>` + pkgDocStyle + `</style>
+  <title>` + title + `</title>
   <script type='text/javascript'>
 // httpGET requests a URL for its effects only.
 function httpGET(url) {
@@ -196,6 +215,13 @@ function httpGET(url) {
 	return false; // disable usual <a href=...> behavior
 }
 
+window.onload = () => {
+	// Hook up the navigation selector.
+	document.getElementById('hdr-Selector').onchange = (e) => {
+		window.location.href = e.target.value;
+	};
+};
+
 // Start a GET /hang request. If it ever completes, the server
 // has disconnected. Show a banner in that case.
 {
@@ -209,10 +235,91 @@ function httpGET(url) {
   </script>
 </head>
 <body>
+<header>
 <div id='disconnected'>Gopls server has terminated. Page is inactive.</div>
+<select id='hdr-Selector'>
+<optgroup label="Documentation">
+  <option label="Overview" value="#hdr-Overview"/>
+  <option label="Index" value="#hdr-Index"/>
+  <option label="Constants" value="#hdr-Constants"/>
+  <option label="Variables" value="#hdr-Variables"/>
+  <option label="Functions" value="#hdr-Functions"/>
+  <option label="Types" value="#hdr-Types"/>
+  <option label="Source Files" value="#hdr-SourceFiles"/>
+</optgroup>
 `)
 
-	escape := html.EscapeString
+	// -- header select element --
+
+	// option emits an <option> for the specified symbol.
+	//
+	// recvType is the apparent receiver type, which may
+	// differ from ReceiverNamed(obj.Signature.Recv).Name
+	// for promoted methods.
+	option := func(obj types.Object, recvType string) {
+		// Render functions/methods as "(recv) Method(p1, ..., pN)".
+		fragment := obj.Name()
+
+		// format parameter names (p1, ..., pN)
+		label := obj.Name() // for a type
+		if fn, ok := obj.(*types.Func); ok {
+			var buf strings.Builder
+			sig := fn.Type().(*types.Signature)
+			if sig.Recv() != nil {
+				fmt.Fprintf(&buf, "(%s) ", sig.Recv().Name())
+				fragment = recvType + "." + fn.Name()
+			}
+			fmt.Fprintf(&buf, "%s(", fn.Name())
+			for i := 0; i < sig.Params().Len(); i++ {
+				if i > 0 {
+					buf.WriteString(", ")
+				}
+				name := sig.Params().At(i).Name()
+				if name == "" {
+					name = "_"
+				}
+				buf.WriteString(name)
+			}
+			buf.WriteByte(')')
+			label = buf.String()
+		}
+
+		fmt.Fprintf(&buf, "  <option label='%s' value='#%s'/>\n", label, fragment)
+	}
+
+	// index of functions
+	fmt.Fprintf(&buf, "<optgroup label='Functions'>\n")
+	for _, fn := range docpkg.Funcs {
+		option(scope.Lookup(fn.Name), "")
+	}
+	fmt.Fprintf(&buf, "</optgroup>\n")
+
+	// index of types
+	fmt.Fprintf(&buf, "<optgroup label='Types'>\n")
+	for _, doctype := range docpkg.Types {
+		option(scope.Lookup(doctype.Name), "")
+	}
+	fmt.Fprintf(&buf, "</optgroup>\n")
+
+	// index of constructors and methods of each type
+	for _, doctype := range docpkg.Types {
+		tname := scope.Lookup(doctype.Name).(*types.TypeName)
+		if len(doctype.Funcs)+len(doctype.Methods) > 0 {
+			fmt.Fprintf(&buf, "<optgroup label='type %s'>\n", doctype.Name)
+			for _, docfn := range doctype.Funcs {
+				option(scope.Lookup(docfn.Name), "")
+			}
+			for _, docmethod := range doctype.Methods {
+				method, _, _ := types.LookupFieldOrMethod(tname.Type(), true, tname.Pkg(), docmethod.Name)
+				option(method, doctype.Name)
+			}
+			fmt.Fprintf(&buf, "</optgroup>\n")
+		}
+	}
+	fmt.Fprintf(&buf, "</select>\n")
+	fmt.Fprintf(&buf, "</header>\n")
+
+	// -- main element --
 
 	// sourceLink returns HTML for a link to open a file in the client editor.
 	sourceLink := func(text, url string) string {
@@ -222,7 +329,7 @@ function httpGET(url) {
 		// We keep the href attribute as it causes the <a> to render
 		// as a link: blue, underlined, with URL hover information.
 		return fmt.Sprintf(`<a href="%[1]s" onclick='return httpGET("%[1]s")'>%[2]s</a>`,
-			escape(url), text)
+			escape(url), escape(text))
 	}
 
 	// objHTML returns HTML for obj.Name(), possibly as a link.
@@ -347,16 +454,79 @@ function httpGET(url) {
 		return escape(buf.String())
 	}
 
-	// pkgRelative qualifies types by package name alone
-	pkgRelative := func(other *types.Package) string {
-		if pkg.Types() == other {
-			return "" // same package; unqualified
+	// fnString is like fn.String() except that it:
+	// - shows the receiver name;
+	// - uses space "(T) M()" not dot "(T).M()" after receiver;
+	// - doesn't bother with the special case for interface receivers
+	//   since it is unreachable for the methods in go/doc.
+	// - elides parameters after the first three: f(a, b, c, ...).
+	fnString := func(fn *types.Func) string {
+		// pkgRelative qualifies types by package name alone
+		pkgRelative := func(other *types.Package) string {
+			if pkg.Types() == other {
+				return "" // same package; unqualified
+			}
+			return other.Name()
 		}
-		return other.Name()
+
+		sig := fn.Type().(*types.Signature)
+
+		// Emit "func (recv T) F".
+		var buf bytes.Buffer
+		buf.WriteString("func ")
+		if recv := sig.Recv(); recv != nil {
+			buf.WriteByte('(')
+			if recv.Name() != "" {
+				buf.WriteString(recv.Name())
+				buf.WriteByte(' ')
+			}
+			types.WriteType(&buf, recv.Type(), pkgRelative)
+			buf.WriteByte(')')
+			buf.WriteByte(' ') // (ObjectString uses a '.' here)
+		} else if pkg := fn.Pkg(); pkg != nil {
+			if s := pkgRelative(pkg); s != "" {
+				buf.WriteString(s)
+				buf.WriteByte('.')
+			}
+		}
+		buf.WriteString(fn.Name())
+
+		// Emit signature.
+		//
+		// Elide parameters after the third one.
+		// WriteSignature is too complex to fork, so we replace
+		// parameters 4+ with "invalid type", format,
+		// then post-process the string.
+		if sig.Params().Len() > 3 {
+
+			// Clone each TypeParam as NewSignatureType modifies them (#67294).
+			cloneTparams := func(seq *types.TypeParamList) []*types.TypeParam {
+				slice := make([]*types.TypeParam, seq.Len())
+				for i := range slice {
+					tparam := seq.At(i)
+					slice[i] = types.NewTypeParam(tparam.Obj(), tparam.Constraint())
+				}
+				return slice
+			}
+
+			sig = types.NewSignatureType(
+				sig.Recv(),
+				cloneTparams(sig.RecvTypeParams()),
+				cloneTparams(sig.TypeParams()),
+				types.NewTuple(append(
+					typesSeqToSlice[*types.Var](sig.Params())[:3],
+					types.NewVar(0, nil, "", types.Typ[types.Invalid]))...),
+				sig.Results(),
+				false) // any final ...T parameter is truncated
+		}
+		types.WriteSignature(&buf, sig, pkgRelative)
+		return strings.ReplaceAll(buf.String(), ", invalid type)", ", ...)")
 	}
 
+	fmt.Fprintf(&buf, "<main>\n")
+
 	// package name
-	fmt.Fprintf(&buf, "<h1>Package %s</h1>\n", pkg.Types().Name())
+	fmt.Fprintf(&buf, "<h1 id='hdr-Overview'>Package %s</h1>\n", pkg.Types().Name())
 
 	// import path
 	fmt.Fprintf(&buf, "<pre class='code'>import %q</pre>\n", pkg.Types().Path())
@@ -369,7 +539,7 @@ function httpGET(url) {
 	fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(docpkg.Doc))
 
 	// symbol index
-	fmt.Fprintf(&buf, "<h2>Index</h2>\n")
+	fmt.Fprintf(&buf, "<h2 id='hdr-Index'>Index</h2>\n")
 	fmt.Fprintf(&buf, "<ul>\n")
 	if len(docpkg.Consts) > 0 {
 		fmt.Fprintf(&buf, "<li><a href='#hdr-Constants'>Constants</a></li>\n")
@@ -377,12 +547,10 @@ function httpGET(url) {
 	if len(docpkg.Vars) > 0 {
 		fmt.Fprintf(&buf, "<li><a href='#hdr-Variables'>Variables</a></li>\n")
 	}
-	scope := pkg.Types().Scope()
 	for _, fn := range docpkg.Funcs {
 		obj := scope.Lookup(fn.Name).(*types.Func)
 		fmt.Fprintf(&buf, "<li><a href='#%s'>%s</a></li>\n",
-			obj.Name(),
-			escape(types.ObjectString(obj, pkgRelative)))
+			obj.Name(), escape(fnString(obj)))
 	}
 	for _, doctype := range docpkg.Types {
 		tname := scope.Lookup(doctype.Name).(*types.TypeName)
@@ -396,19 +564,15 @@ function httpGET(url) {
 			for _, docfn := range doctype.Funcs {
 				obj := scope.Lookup(docfn.Name).(*types.Func)
 				fmt.Fprintf(&buf, "<li><a href='#%s'>%s</a></li>\n",
-					docfn.Name,
-					escape(types.ObjectString(obj, pkgRelative)))
+					docfn.Name, escape(fnString(obj)))
 			}
 			// methods
 			for _, docmethod := range doctype.Methods {
 				method, _, _ := types.LookupFieldOrMethod(tname.Type(), true, tname.Pkg(), docmethod.Name)
-				// TODO(adonovan): style: change the . into a space in
-				// ObjectString's "func (T).M()", and hide unexported
-				// embedded types.
 				fmt.Fprintf(&buf, "<li><a href='#%s.%s'>%s</a></li>\n",
 					doctype.Name,
 					docmethod.Name,
-					escape(types.ObjectString(method, pkgRelative)))
+					escape(fnString(method.(*types.Func))))
 			}
 			fmt.Fprintf(&buf, "</ul>\n")
 		}
@@ -447,7 +611,7 @@ function httpGET(url) {
 	}
 
 	// package-level functions
-	fmt.Fprintf(&buf, "<h2>Functions</h2>\n")
+	fmt.Fprintf(&buf, "<h2 id='hdr-Functions'>Functions</h2>\n")
 	// funcs emits a list of package-level functions,
 	// possibly organized beneath the type they construct.
 	funcs := func(funcs []*doc.Func) {
@@ -467,7 +631,7 @@ function httpGET(url) {
 	funcs(docpkg.Funcs)
 
 	// types and their subelements
-	fmt.Fprintf(&buf, "<h2>Types</h2>\n")
+	fmt.Fprintf(&buf, "<h2 id='hdr-Types'>Types</h2>\n")
 	for _, doctype := range docpkg.Types {
 		tname := scope.Lookup(doctype.Name).(*types.TypeName)
 
@@ -506,15 +670,35 @@ function httpGET(url) {
 	}
 
 	// source files
-	fmt.Fprintf(&buf, "<h2>Source files</h2>\n")
+	fmt.Fprintf(&buf, "<h2 id='hdr-SourceFiles'>Source files</h2>\n")
 	for _, filename := range docpkg.Filenames {
 		fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n",
 			sourceLink(filepath.Base(filename), posURL(filename, 1, 1)))
 	}
 
+	fmt.Fprintf(&buf, "</main>\n")
+	fmt.Fprintf(&buf, "</body>\n")
+	fmt.Fprintf(&buf, "</html>\n")
+
 	return buf.Bytes(), nil
 }
 
+// typesSeq abstracts various go/types sequence types:
+// MethodSet, Tuple, TypeParamList, TypeList.
+// TODO(adonovan): replace with go1.23 iterators.
+type typesSeq[T any] interface {
+	Len() int
+	At(int) T
+}
+
+func typesSeqToSlice[T any](seq typesSeq[T]) []T {
+	slice := make([]T, seq.Len())
+	for i := range slice {
+		slice[i] = seq.At(i)
+	}
+	return slice
+}
+
 // (partly taken from pkgsite's typography.css)
 const pkgDocStyle = `
 body {
@@ -621,6 +805,21 @@ a:hover > * {
 
 #pkgsite { height: 1.5em; }
 
+header {
+  position: sticky;
+  top: 0;
+  left: 0;
+  width: 100%;
+  padding: 0.3em;
+}
+
+#hdr-Selector {
+  margin-right: 0.3em;
+  float: right;
+  min-width: 25em;
+  padding: 0.3em;
+}
+
 #disconnected {
   position: fixed;
   top: 1em;
